#include <unistd.h>
#include <sys/wait.h>
#include <exception>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <memory>
#include <vector>
#include <cstdio>
#include <cstring>
#include <cstdarg>
#include <cstdlib>
#include <cerrno>
using namespace std;

class spawn {
	int write_pipe[2];
	int read_pipe[2];
	int child_pid;
public:
	spawn(char * const argv[]) {
		if (pipe(write_pipe) == -1) throw runtime_error("Failed to create pipe");
		if (pipe(read_pipe) == -1) throw runtime_error("Failed to create pipe");
		child_pid = fork();
		if (child_pid == -1) throw runtime_error("Failed to fork process");
		if (child_pid == 0) { // in child
			if (dup2(write_pipe[0], STDIN_FILENO) == -1) throw runtime_error("Failed to dup handle");
			if (dup2(read_pipe[1], STDOUT_FILENO) == -1) throw runtime_error("Failed to dup handle");
			if (close(write_pipe[0]) == -1) throw runtime_error("Failed to close handle");
			if (close(write_pipe[1]) == -1) throw runtime_error("Failed to close handle");
			if (close(read_pipe[0]) == -1) throw runtime_error("Failed to close handle");
			if (close(read_pipe[1]) == -1) throw runtime_error("Failed to close handle");
			if (execv(argv[0], argv) == -1) {
				cerr << "Error: Failed to launch program\n";
				exit(1);
			}
		} else { // in parent
			if (close(write_pipe[0]) == -1) throw runtime_error("Failed to close handle");
			if (close(read_pipe[1]) == -1) throw runtime_error("Failed to close handle");
			signal(SIGPIPE, SIG_IGN); // 
		}
	}

	bool send_line(const char *fmt, ...) {
		char buf[2000];
		va_list ap;
		va_start(ap, fmt);
		//int n = vsnprintf(0, 0, fmt, ap);
		//if (n < 0)  throw runtime_error("Invalid format");
		//if (n >= sizeof(buf)) throw runtime_error("Buffer overflow");
		int n = vsnprintf(buf, sizeof(buf), fmt, ap);
		va_end(ap);
		return write(write_pipe[1], buf, n) == n;
	}

	bool send_eof() {
		return close(write_pipe[1]) != -1;
	}

	bool recv_line(char buf[], int size) {
		for (int n = 0; n + 1 < size; ++n) {
			int m = read(read_pipe[0], buf + n, 1);
			if (m < 0) {
				return false;
			}
			if (m == 0) {
				buf[n] = 0;
				return true;
			}
			if (buf[n] == '\n') {
				buf[n] = 0;
				return true;
			}
		}
		buf[size - 1] = 0;
		return true;
	}

	int wait() {
		int status;
		waitpid(child_pid, &status, 0);
		return status;
	}
};

#define done(x, y) ( \
	fo << (x) << '\n' << (y) << '\n', \
	cerr << "Res: " << (x) << '\n' << "Msg: " << (y) << '\n')

int main(int argc, char *argv[]) {
	if (argc < 4) {
		cerr << "Usage: " << argv[0] << " input output child [params ...]'n";
		return 0;
	}

	ifstream fi(argv[1]);
	ofstream fo(argv[2]);
	fo << fixed << setprecision(20);

	spawn *prog[4];
	for (int i = 0; i < 4; ++i) {
		prog[i] = new spawn(argv + 3);
	}

	int b, m, t, n;
	long double score_mult;
	fi >> b >> m >> score_mult >> t >> n;
	score_mult = 1 / score_mult;

	bool good_com = true;
	for (int i = 0; i < 4; ++i) {
		if (!prog[i]->send_line("%d\n%d %d\n", i >= 2, b, m)) {
			cerr << i << " initialization failed " << errno << '\n';
			done(0, "Presentation error");
			good_com = false;
		}
		cerr << i << " <- " << (i >= 2) << ' ' << b << ' ' << m << '\n';
	}

	vector<char *> message(n);
	vector<int> encode_id(n), decode_id(n);
	for(int i = 0; i < n; ++i) {
		message[i] = new char[m + 1];
	}
	vector<char *> mem(n + 1);
	for(int i = 0; i <= n; ++i) {
		mem[i] = new char[b + 1];
	}
	memset(mem[0], '0', b);
	mem[0][b] = 0;
	char buffer[2000], junk[10];
	bool n_reached = false;
	int write_cnt = 0;
	for (int ct = 0; ct < t && good_com; ++ct) {
		for (int i = 0; i < n; ++i) {
			fi >> message[i] >> encode_id[i] >> decode_id[i];
		}
		int cur_write_cnt = 0;
		for (int i = 0; i < n && good_com; ++i) {
			int id = encode_id[i];
			if (!prog[id]->send_line("%d\n%s %s\n", 1, mem[i], message[i])) {
				cerr << id << " sending write request failed " << errno << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			cerr << id << " <- 1 " << mem[i] << ' ' << message[i] << '\n';
			if (!prog[id]->recv_line(buffer, sizeof(buffer))) {
				cerr << id << " reading write status failed " << errno << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			int suc;
			if (sscanf(buffer, "%d%5s", &suc, junk) != 1 || (suc != 1 && suc != 0)) {
				cerr << id << " -> " << buffer << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			if (suc == 0) {
				cerr << id << " -> 0" << '\n';
				break;
			}
			if (!prog[id]->recv_line(buffer, sizeof(buffer))) {
				cerr << id << " reading write result failed " << errno << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			int end = strspn(buffer, "01");
			if (end != b || sscanf(buffer + end, "%5s", junk) != EOF) {
				cerr << id << " -> 1 " << buffer << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			buffer[end] = 0;
			cerr << id << " -> 1 " << buffer << '\n';
			for (int j = 0; j < b; ++j) {
				if (buffer[j] < mem[i][j]) {
					cerr << id << " writer decreased bit " << j << " of memory\n";
					done(0, "Invalid write");
					good_com = false;
					break;
				}
			}
			++cur_write_cnt;
			++write_cnt;
			strcpy(mem[i + 1], buffer);
		}
		if (cur_write_cnt == n && good_com) {
			n_reached = true;
		}
		for (int i = cur_write_cnt; i > 0 && good_com; --i) {
			int id = 2 + decode_id[i - 1];
			if (!prog[id]->send_line("%d\n%s\n", 1, mem[i])) {
				cerr << id << " sending read request failed " << errno << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			cerr << id << " <- 1 " << mem[i] << '\n';
			if (!prog[id]->recv_line(buffer, sizeof(buffer))) {
				cerr << id << " reading read result failed " << errno << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			int end = strspn(buffer, "01");
			if (end != m || sscanf(buffer + end, "%5s", junk) != EOF) {
				cerr << id << " -> 1 " << buffer << '\n';
				done(0, "Presentation error");
				good_com = false;
				break;
			}
			buffer[end] = 0;
			cerr << id << " -> " << buffer << '\n';
			for (int j = 0; j < m; ++j) {
				if (buffer[j] != message[i - 1][j]) {
					cerr << id << " reader misread bit " << j << " of value\n";
					done(-1, "Incorrect read");
					good_com = false;
					break;
				}
			}
		}
	}
	for (int i = 0; i < 4; ++i) {
		if (!prog[i]->send_line("%d\n", 0)) {
			cerr << i << " shutdown failed " << errno << '\n';
			done(0, "Presentation error");
			good_com = false;
		}
		prog[i]->send_eof();
		cerr << i << " <- " << 0 << '\n';
	}
	int status = 0;
	for (int i = 0; i < 4; ++i) {
		int cur_status = prog[i]->wait();
		if (cur_status) {
			status = cur_status;
		}
		cerr << i << " wait " << cur_status << '\n';
	}
	if (good_com){
		long double res = score_mult * write_cnt / t;
		if (n_reached){
			done(res, "Output is correct, congratulations: you wrote more messages than was thought posssible");
		} else if (write_cnt * b > 120000){
			done(res, "Output is correct, congratulations: you reached BN > 120000");
		} else {
			done(res, "Output is correct");
		}
	}
	return status; // pass child's status back to environment
}
